==> fap-0.9.2.tar.bz2: fap-0.9.2/src/audio/audio.cpp <==
/*
 *  Copyright (C) 2006-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "audio.h"
#include "inputTypes.h"
#include "outputFactory.h"
#include "settings.h"
#include "input.h"
#include "output.h"
#include "converterFactory.h"

#define DECIMALS 10
//#define PROFILE

#ifdef PROFILE

#if FOX_MINOR<7 || (FOX_MINOR=7 && FOX_LEVEL<10)
namespace FX {
extern FXAPI FXlong fxgetticks();
}
#endif

#  define PROFILE_START	FXlong ticks=FX::fxgetticks();
#  define PROFILE_END	ticks=FX::fxgetticks()-ticks; \
			fxmessage("Ticks %d\n", (int)ticks);
#else
#  define PROFILE_START
#  define PROFILE_END
#endif

/*****************************************************************/

FXint audioThread::run() {

	switch (_audio->outputPrecision()) {
	case U8:
		_audio->loop<FXuchar>();
		break;
	case S16:
		_audio->loop<FXshort>();
		break;
	case S24:
	case S32:
		fxmessage("Not supported yet\n");
		break;
	}

	return 1;
}

/*****************************************************************/

inline sample_t audio::outputPrecision() {

	switch (_input->precision()) {
	case U8:
	case S16:
	case S24:
	case S32:
		return _input->precision();
	case SAMPLE_FLOAT:
	case SAMPLE_FIXED:
		switch (SETTINGS->bits()) {
		case 8:
			return U8;
		case 16:
			return S16;
		}
	}
}

template<> inline void audio::elaborate<FXuchar>(size_t size) {

#if defined HAVE_BS2B && BS2B_VERSION_MAJOR == 3
	if (_bs2bdp) {
		uint8_t *buf=(uint8_t*)_output->buffer();
		bs2b_cross_feed_u8(_bs2bdp, buf, size/2);
	}
#endif
}

template<> inline void audio::elaborate<FXshort>(size_t size) {

#if (FOX_BIGENDIAN == 1)
	{
		//Swap bytes on big endian machines
		FXshort *buf=(FXshort*)_output->buffer();
		const FXshort *end=buf+(size/2);
		do {
			const FXShort tmp=*buf;
			*buf++=((tmp&0x00FF)<<8)&((tmp&0xFF00)>>8);
		} while (buf<end);
	}
#endif
#ifdef HAVE_BS2B
	if (_bs2bdp) {
#if BS2B_VERSION_MAJOR == 3
		int16_t *buf=(int16_t*)_output->buffer();
		bs2b_cross_feed_s16le(_bs2bdp, buf, size/4);
#else
		short *buf=(short*)_output->buffer();
		const short *end=buf+(size/2);
		do {
			bs2b_cross_feed_16(_bs2bdp, buf);
			buf+=2;
		} while (buf<end);
#endif
	}
#endif
}

template<typename T>
void audio::loop() {

	FXTRACE((10, "loop %d bit\n", sizeof(T)<<3));
	while (_playing) {
PROFILE_START
		size_t size=_input->fillBuffer(
			_converter?_converter->buffer():_output->buffer(),
			_converter?_converter->bufSize():_bufferSize,
			_seconds);

		if (!size) {
			_guiMsg->setData((void*)SONG_ENDED);
			_guiMsg->signal();
			if (_preload) {
				_input=_preload;
				_preload=0;
				_seconds=0;
				continue;
			} else
				break;
		}

		if (_converter)
			size=_converter->convert(_output->buffer(), _bufferSize);

		elaborate<T>(size);
PROFILE_END

		if (!_output->write(_output->buffer(), _bufferSize) && _playing) {
			fxmessage("Output error\n");
			_guiMsg->setData((void*)OUTPUT_ERROR);
			_guiMsg->signal();
			_playing=false;
			break;
		}

		_buffers-=(1<<DECIMALS);
		if (_buffers<0) {
			do {
				_buffers+=_bufPerSec;
				_seconds++;
			} while (_buffers<0);
			_guiMsg->setData((_seconds!=_input->time()-5)?
				(void*)UPD_TIME:(void*)PRELOAD_SONG);
			_guiMsg->signal();
		}
	}
}

/*****************************************************************/

audio::audio(FXObject *o, FXSelector guiMsg) :
	_input(0),
	_preload(0),
	_state(STOP),
	_playing(false),
	_seconds(0)
#ifdef HAVE_BS2B
	,_bs2bdp(0)
#endif
	{

	_guiMsg=new FXGUISignal(FXApp::instance(), o, guiMsg);

	_volume=FXApp::instance()->reg().readIntEntry("Audio Settings", "volume", 50);

	FXTRACE((10, "apiName:%s\n", SETTINGS->apiName()));
	_output=OFACTORY->get(SETTINGS->apiName());
}

audio::~audio() {

	stop();

	FXApp::instance()->reg().writeIntEntry("Audio Settings", "volume", _volume);

	delete _output;
	delete _guiMsg;
}

bool audio::play(input* i, int pos) {

	if ((!_output) || (!i->songLoaded()) || (_state==PLAY))
		return false;

	FXTRACE((10, "audio::play\n"));

	_input=i;

	unsigned int _card=0;
	const char* card=SETTINGS->card();
	for (int i=0; i<_output->devices(); i++) {
		if (!compare(card, _output->device(i))) {
			_card=i;
			break;
		}
	}

	unsigned int sampleRate=_input->samplerate();
	//unsigned int sampleRate=11025;

	// FLOAT and FIXED are 16bit for now
	const unsigned int precision=(outputPrecision()==U8)?1:2; //FIXME
	FXTRACE((10, "Setting parameters %d:%d:%d\n", sampleRate, _input->channels(), precision));
	_bufferSize=_output->open(_card, sampleRate, _input->channels(), precision);
	if (!_bufferSize)
		return false;

	FXTRACE((10,"Output samplerate %d\n", sampleRate));

	// Check if soundcard supports requested samplerate
	_converter=CFACTORY->get(_input->samplerate(), sampleRate, _bufferSize,
		_input->channels(), _input->precision(), outputPrecision(), _input->fract());

	_bufPerSec=((sampleRate<<DECIMALS)*_input->channels()*precision)/_bufferSize;
	_buffers=_bufPerSec;
#ifdef HAVE_BS2B
	if (SETTINGS->bs2b() && _input->channels()==2
#if BS2B_VERSION_MAJOR == 2
		&& outputPrecision()==S16
#endif
	) {
		_bs2bdp=bs2b_open();
		if (_bs2bdp) {
			FXTRACE((10, "bs2b enabled\n"));
			bs2b_set_srate(_bs2bdp, sampleRate);
			bs2b_set_level(_bs2bdp, BS2B_DEFAULT_CLEVEL);
		}
	}
#endif
	_output->volume(_volume);

	_input->seek(pos);

	_thread=new audioThread(this);

	_playing=true;
	_state=PLAY;
	_thread->start();

	return true;
}

void audio::pause() {

	switch (_state) {
	case PLAY:
		FXTRACE((10, "Pause\n"));
		_playing=false;
		_output->pause();
		_thread->join();
		_state=PAUSE;
		break;
	case PAUSE:
		FXTRACE((10, "Unpause\n"));
		_playing=true;
		_thread->start();
		_state=PLAY;
		break;
	}
}

bool audio::stop() {

	if ((!_output) || (_state==STOP))
		return false;

	FXTRACE((10, "audio::stop\n"));

	_playing=false;

	_output->stop();

	_thread->join();

	_output->close();

	delete _thread;

	_seconds=0;
	_state=STOP;
#ifdef HAVE_BS2B
	if (_bs2bdp) {
		bs2b_close(_bs2bdp);
		_bs2bdp=0;
	}
#endif
	if (_converter)
		delete _converter;

	return true;
}

void audio::volume(const int vol) {

	_volume=vol;
	if (_output)
		_output->volume(_volume);
}

bool audio::gapless(input* const i) {

	if ((_input->samplerate()==i->samplerate()) &&
	(_input->channels()==i->channels()) &&
	(_input->precision()==i->precision())) {
		_preload=i;
		return true;
	} else {
		_preload=0;
		return false;
	}
}

void audio::setOpts() {

	delete _output;
	_output=OFACTORY->get(SETTINGS->apiName());
}

/*****************************************************************/

FXDEFMAP(audioConfig) audioConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_API,		audioConfig::onCmdApi),
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_CARD,		audioConfig::onCmdCard),
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_BITS,		audioConfig::onCmdBits),
};

FXIMPLEMENT(audioConfig, configFrame, audioConfigMap, ARRAYNUMBER(audioConfigMap))

audioConfig::audioConfig(FXComposite* win) :
	configFrame(win, "Audio") {

	new FXLabel(matrix(), tr("API"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	_apiList=new FXListBox(matrix(), this, audioConfig::ID_API, LAYOUT_FILL|FRAME_NORMAL);

	new FXLabel(matrix(), tr("Card"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	_cardList=new FXListBox(matrix(), this, audioConfig::ID_CARD, LAYOUT_FILL|FRAME_NORMAL);

	new FXLabel(matrix(), tr("Default bitdepth"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *bitBox=new FXListBox(matrix(), this, audioConfig::ID_BITS, LAYOUT_FILL|FRAME_NORMAL);
	bitBox->fillItems("8\n16");
	bitBox->setNumVisible(2);
	unsigned int val;
	switch (SETTINGS->bits()) {
	case 8:
		val=0;
		break;
	default:
	case 16:
		val=1;
		break;
	}
	bitBox->setCurrentItem(val);

	const int apis=OFACTORY->num();
	for (int i=0; i<apis; i++)
		_apiList->appendItem(OFACTORY->name(i));

	_apiList->setNumVisible(apis);
	_apiList->setCurrentItem(_apiList->findItem(SETTINGS->apiName()));

	setCards(OFACTORY->get(SETTINGS->apiName()));
}

long audioConfig::onCmdApi(FXObject*, FXSelector, void* ptr) {

	FXTRACE((10, "onCmdApi\n"));

	const FXint val=(FXint)(FXival)ptr;
	setCards(OFACTORY->get(val));

	SETTINGS->_apiName=_apiList->getItemText(val);

	return 1;
}

long audioConfig::onCmdCard(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;

	SETTINGS->_card=_cardList->getItemText(val);

	return 1;
}

void audioConfig::setCards(const output* const audio) {

	if (!audio)
		return;

	_cardList->clearItems();

	const int devices=audio->devices();
	FXTRACE((10, "Devices: %d\n", devices));
	for (int i=0; i<devices; i++) {
		_cardList->appendItem(audio->device(i));
	}

	delete audio;

	_cardList->setNumVisible((devices>5)?5:devices);

	FXint val=_cardList->findItem(SETTINGS->card());
	if (val>=0)
		_cardList->setCurrentItem(val);
}

long audioConfig::onCmdBits(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS->_bits=8;
		break;
	case 1:
		SETTINGS->_bits=16;
		break;
	}
	return 1;
}

==> fap-0.10.0.tar.bz2: fap-0.10.0/src/audio/audio.cpp <==
/*
 *  Copyright (C) 2006-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "audio.h"
#include "inputTypes.h"
#include "outputFactory.h"
#include "settings.h"
#include "input.h"
#include "output.h"
#include "converterFactory.h"

#define DECIMALS 10
//#define PROFILE

#ifdef PROFILE

#if FOX_MINOR<7 || (FOX_MINOR=7 && FOX_LEVEL<10)
namespace FX {
extern FXAPI FXlong fxgetticks();
}
#endif

#  define PROFILE_START	FXlong ticks=FX::fxgetticks();
#  define PROFILE_END	ticks=FX::fxgetticks()-ticks; \
			fxmessage("Ticks %d\n", (int)ticks);
#else
#  define PROFILE_START
#  define PROFILE_END
#endif

/*****************************************************************/

FXint audioThread::run() {

	switch (_audio->outputPrecision()) {
	case U8:
		_audio->loop<FXuchar>();
		break;
	case S16:
		_audio->loop<FXshort>();
		break;
	case S24:
	case S32:
		fxmessage("Not supported yet\n");
		break;
	}

	return 1;
}

/*****************************************************************/

inline sample_t audio::outputPrecision() {

	switch (_input->precision()) {
	case U8:
	case S16:
	case S24:
	case S32:
		return _input->precision();
	case SAMPLE_FLOAT:
	case SAMPLE_FIXED:
		switch (SETTINGS->bits()) {
		case 8:
			return U8;
		case 16:
			return S16;
		}
	}
}

template<> inline void audio::process<FXuchar>(size_t size) {

#if defined HAVE_BS2B && BS2B_VERSION_MAJOR == 3
	if (_bs2bdp) {
		uint8_t *buf=(uint8_t*)_output->buffer();
		bs2b_cross_feed_u8(_bs2bdp, buf, size/2);
	}
#endif
}

template<> inline void audio::process<FXshort>(size_t size) {

#if (FOX_BIGENDIAN == 1)
	{
		//Swap bytes on big endian machines
		FXshort *buf=(FXshort*)_output->buffer();
		const FXshort *end=buf+(size/2);
		do {
			const FXShort tmp=*buf;
			*buf++=((tmp&0x00FF)<<8)&((tmp&0xFF00)>>8);
		} while (buf<end);
	}
#endif
#ifdef HAVE_BS2B
	if (_bs2bdp) {
#if BS2B_VERSION_MAJOR == 3
		int16_t *buf=(int16_t*)_output->buffer();
		bs2b_cross_feed_s16le(_bs2bdp, buf, size/4);
#else
		short *buf=(short*)_output->buffer();
		const short *end=buf+(size/2);
		do {
			bs2b_cross_feed_16(_bs2bdp, buf);
			buf+=2;
		} while (buf<end);
#endif
	}
#endif
}

template<typename T>
void audio::loop() {

	FXTRACE((10, "loop %d bit\n", sizeof(T)<<3));
	while (_playing) {
PROFILE_START
		size_t size=_input->fillBuffer(
			_converter?_converter->buffer():_output->buffer(),
			_converter?_converter->bufSize():_bufferSize,
			_seconds);

		if (!size) {
			_guiMsg->setData((void*)SONG_ENDED);
			_guiMsg->signal();
			if (_preload) {
				_input=_preload;
				_preload=0;
				_seconds=0;
				continue;
			} else
				break;
		}

		if (_converter)
			size=_converter->convert(_output->buffer(), _bufferSize);

		process<T>(size);
PROFILE_END

		if (!_output->write(_output->buffer(), _bufferSize) && _playing) {
			fxmessage("Output error\n");
			_guiMsg->setData((void*)OUTPUT_ERROR);
			_guiMsg->signal();
			_playing=false;
			break;
		}

		_buffers-=(1<<DECIMALS);
		if (_buffers<0) {
			do {
				_buffers+=_bufPerSec;
				_seconds++;
			} while (_buffers<0);
			_guiMsg->setData((_seconds!=_input->time()-5)?
				(void*)UPD_TIME:(void*)PRELOAD_SONG);
			_guiMsg->signal();
		}
	}
}

/*****************************************************************/

audio::audio(FXObject *o, FXSelector guiMsg) :
	_input(0),
	_preload(0),
	_state(STOP),
	_playing(false),
	_seconds(0)
#ifdef HAVE_BS2B
	,_bs2bdp(0)
#endif
	{

	_guiMsg=new FXGUISignal(FXApp::instance(), o, guiMsg);

	_volume=FXApp::instance()->reg().readIntEntry("Audio Settings", "volume", 50);

	FXTRACE((10, "apiName:%s\n", SETTINGS->apiName()));
	_output=OFACTORY->get(SETTINGS->apiName());
}

audio::~audio() {

	stop();

	FXApp::instance()->reg().writeIntEntry("Audio Settings", "volume", _volume);

	delete _output;
	delete _guiMsg;
}

bool audio::play(input* i, int pos) {

	if ((!_output) || (!i->songLoaded()) || (_state==PLAY))
		return false;

	FXTRACE((10, "audio::play\n"));

	_input=i;

	unsigned int _card=0;
	const char* card=SETTINGS->card();
	for (int i=0; i<_output->devices(); i++) {
		if (!compare(card, _output->device(i))) {
			_card=i;
			break;
		}
	}

	unsigned int sampleRate=_input->samplerate();
	//unsigned int sampleRate=11025;

	// FIXME only supports 8/16 bits
	const unsigned int precision=(outputPrecision()==U8)?1:2;
	FXTRACE((10, "Setting parameters %d:%d:%d\n", sampleRate, _input->channels(), precision));
	_bufferSize=_output->open(_card, sampleRate, _input->channels(), precision);
	if (!_bufferSize)
		return false;

	FXTRACE((10,"Output samplerate %d\n", sampleRate));

	// Check if soundcard supports requested samplerate
	_converter=CFACTORY->get(_input->samplerate(), sampleRate, _bufferSize,
		_input->channels(), _input->precision(), outputPrecision(), _input->fract());

	_bufPerSec=((sampleRate<<DECIMALS)*_input->channels()*precision)/_bufferSize;
	_buffers=_bufPerSec;
#ifdef HAVE_BS2B
	if (SETTINGS->bs2b() && _input->channels()==2
#if BS2B_VERSION_MAJOR == 2
		&& outputPrecision()==S16
#endif
	) {
		_bs2bdp=bs2b_open();
		if (_bs2bdp) {
			FXTRACE((10, "bs2b enabled\n"));
			bs2b_set_srate(_bs2bdp, sampleRate);
			bs2b_set_level(_bs2bdp, BS2B_DEFAULT_CLEVEL);
		}
	}
#endif
	_output->volume(_volume);

	_input->seek(pos);

	_thread=new audioThread(this);

	_playing=true;
	_state=PLAY;
	_thread->start();

	return true;
}

void audio::pause() {

	switch (_state) {
	case PLAY:
		FXTRACE((10, "Pause\n"));
		_playing=false;
		_output->pause();
		_thread->join();
		_state=PAUSE;
		break;
	case PAUSE:
		FXTRACE((10, "Unpause\n"));
		_playing=true;
		_thread->start();
		_state=PLAY;
		break;
	}
}

bool audio::stop() {

	if ((!_output) || (_state==STOP))
		return false;

	FXTRACE((10, "audio::stop\n"));

	_playing=false;

	_output->stop();

	_thread->join();

	_output->close();

	delete _thread;

	_seconds=0;
	_state=STOP;
#ifdef HAVE_BS2B
	if (_bs2bdp) {
		bs2b_close(_bs2bdp);
		_bs2bdp=0;
	}
#endif
	if (_converter)
		delete _converter;

	return true;
}

void audio::volume(const int vol) {

	_volume=vol;
	if (_output)
		_output->volume(_volume);
}

bool audio::gapless(input* const i) {

	if ((_input->samplerate()==i->samplerate()) &&
	(_input->channels()==i->channels()) &&
	(_input->precision()==i->precision())) {
		_preload=i;
		return true;
	} else {
		_preload=0;
		return false;
	}
}

void audio::setOpts() {

	delete _output;
	_output=OFACTORY->get(SETTINGS->apiName());
}

/*****************************************************************/

FXDEFMAP(audioConfig) audioConfigMap[] = {
	// Message_Type		ID				Message_Handler
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_API,		audioConfig::onCmdApi),
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_CARD,		audioConfig::onCmdCard),
	FXMAPFUNC(SEL_COMMAND,	audioConfig::ID_BITS,		audioConfig::onCmdBits),
};

FXIMPLEMENT(audioConfig, configFrame, audioConfigMap, ARRAYNUMBER(audioConfigMap))

audioConfig::audioConfig(FXComposite* win) :
	configFrame(win, "Audio") {

	new FXLabel(matrix(), tr("API"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	_apiList=new FXListBox(matrix(), this, audioConfig::ID_API, LAYOUT_FILL|FRAME_NORMAL);

	new FXLabel(matrix(), tr("Card"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	_cardList=new FXListBox(matrix(), this, audioConfig::ID_CARD, LAYOUT_FILL|FRAME_NORMAL);

	new FXLabel(matrix(), tr("Default bitdepth"), 0, LAYOUT_FILL|JUSTIFY_RIGHT);
	FXListBox *bitBox=new FXListBox(matrix(), this, audioConfig::ID_BITS, LAYOUT_FILL|FRAME_NORMAL);
	bitBox->fillItems("8\n16");
	bitBox->setNumVisible(2);
	unsigned int val;
	switch (SETTINGS->bits()) {
	case 8:
		val=0;
		break;
	default:
	case 16:
		val=1;
		break;
	}
	bitBox->setCurrentItem(val);

	const int apis=OFACTORY->num();
	for (int i=0; i<apis; i++)
		_apiList->appendItem(OFACTORY->name(i));

	_apiList->setNumVisible(apis);
	_apiList->setCurrentItem(_apiList->findItem(SETTINGS->apiName()));

	setCards(OFACTORY->get(SETTINGS->apiName()));
}

long audioConfig::onCmdApi(FXObject*, FXSelector, void* ptr) {

	FXTRACE((10, "onCmdApi\n"));

	const FXint val=(FXint)(FXival)ptr;
	setCards(OFACTORY->get(val));

	SETTINGS->_apiName=_apiList->getItemText(val);

	return 1;
}

long audioConfig::onCmdCard(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;

	SETTINGS->_card=_cardList->getItemText(val);

	return 1;
}

void audioConfig::setCards(const output* const audio) {

	if (!audio)
		return;

	_cardList->clearItems();

	const int devices=audio->devices();
	FXTRACE((10, "Devices: %d\n", devices));
	for (int i=0; i<devices; i++) {
		_cardList->appendItem(audio->device(i));
	}

	delete audio;

	_cardList->setNumVisible((devices>5)?5:devices);

	FXint val=_cardList->findItem(SETTINGS->card());
	if (val>=0)
		_cardList->setCurrentItem(val);
}

long audioConfig::onCmdBits(FXObject*, FXSelector, void* ptr) {

	const FXint val=(FXint)(FXival)ptr;
	switch (val) {
	case 0:
		SETTINGS->_bits=8;
		break;
	case 1:
		SETTINGS->_bits=16;
		break;
	}
	return 1;
}
